RPC(
  compressed,
  INT32_FIELD(msg_len)
  INT32_FIELD(request_id)
  INT32_FIELD(response_to)
  INT32_FIELD(opcode)
  INT32_FIELD(original_opcode)
  INT32_FIELD(uncompressed_size)
  UINT8_FIELD(compressor_id)
  RAW_BUFFER_FIELD(compressed_message)
)
